@@ -1,6 +1,7 @@
 # /usr/bin/env python3
 
 import datetime
+import glob
 import itertools
 import json
 import logging
@@ -15,11 +16,17 @@ import osxmetadata
 
 from ._version import __version__
 from .attributes import _LONG_NAME_WIDTH, _SHORT_NAME_WIDTH, ATTRIBUTES
+from .backup import load_backup_file, write_backup_file
 from .classes import _AttributeList, _AttributeTagsList
-from .constants import _BACKUP_FILENAME, _TAGS_NAMES
-from .utils import load_backup_file, validate_attribute_value, write_backup_file
+from .constants import (
+    _BACKUP_FILENAME,
+    _COLORNAMES_LOWER,
+    _FINDERINFO_NAMES,
+    _TAGS_NAMES,
+    FINDER_COLOR_NONE,
+)
+from .findertags import Tag, tag_factory
 
-# TODO: add md5 option
 # TODO: how is metadata on symlink handled?
 # should symlink be resolved before gathering metadata?
 # currently, symlinks are resolved before handling metadata but not sure this is the right behavior
@@ -29,15 +36,13 @@ from .utils import load_backup_file, validate_attribute_value, write_backup_file
 # TODO: add selective restore (e.g only restore files matching command line path)
 #   e.g osxmetadata -r meta.json *.pdf
 
-# TODO: need special handling for --set color GREEN, etc.
-
 
 # Click CLI object & context settings
 class CLI_Obj:
     def __init__(self, debug=False, files=None):
         self.debug = debug
         if debug:
-            osxmetadata._set_debug(True)
+            osxmetadata.debug._set_debug(True)
 
         self.files = files
 
@@ -74,7 +79,7 @@ class MyClickCommand(click.Command):
         )
         formatter.write("\n")
         formatter.write_text('For example: --set findercomment "Hello world"')
-        formatter.write_text('or:          --set kMDFinderComment "Hello world"')
+        formatter.write_text('or:          --set kMDItemFinderComment "Hello world"')
         formatter.write_text(
             'or:          --set com.apple.metadata:kMDItemFinderComment "Hello world"'
         )
@@ -89,18 +94,28 @@ class MyClickCommand(click.Command):
         formatter.write_text(
             "Options are executed in the following order regardless of order "
             + "passed on the command line: "
-            + "restore, wipe, clear, set, append, update, remove, mirror, get, list, backup.  "
+            + "restore, wipe, copyfrom, clear, set, append, update, remove, mirror, get, list, backup.  "
             + "--backup and --restore are mutually exclusive.  "
             + "Other options may be combined or chained together."
         )
         formatter.write("\n")
+        formatter.write_text(
+            "Finder tags (tags attribute) contain both a name and an optional color. "
+            + "To specify the color, append comma + color name (e.g. 'red') after the "
+            + "tag name.  For example --set tags Foo,red. "
+            + "Valid color names are: "
+            + f"{', '.join([color for color, colorid in _COLORNAMES_LOWER.items() if colorid != FINDER_COLOR_NONE])}. "
+            + "If color is not specified but a tag of the same name has already been assigned a color "
+            + "in the Finder, the same color will automatically be assigned. "
+        )
+        formatter.write("\n")
 
         formatter.write_dl(attr_tuples)
         help_text += formatter.getvalue()
         return help_text
 
 
-# CTX_SETTINGS = dict(help_option_names=["--help", "-h"])
+# All the command line options defined here
 FILES_ARGUMENT = click.argument(
     "files", metavar="FILE", nargs=-1, type=click.Path(exists=True)
 )
@@ -134,6 +149,7 @@ DEBUG_OPTION = click.option(
 )
 SET_OPTION = click.option(
     "--set",
+    "-s",
     "set_",
     metavar="ATTRIBUTE VALUE",
     help="Set ATTRIBUTE to VALUE.",
@@ -143,6 +159,7 @@ SET_OPTION = click.option(
 )
 GET_OPTION = click.option(
     "--get",
+    "-g",
     help="Get value of ATTRIBUTE.",
     metavar="ATTRIBUTE",
     nargs=1,
@@ -151,6 +168,7 @@ GET_OPTION = click.option(
 )
 LIST_OPTION = click.option(
     "--list",
+    "-l",
     "list_",
     help="List all metadata attributes for FILE.",
     is_flag=True,
@@ -158,6 +176,7 @@ LIST_OPTION = click.option(
 )
 CLEAR_OPTION = click.option(
     "--clear",
+    "-c",
     help="Remove attribute from FILE.",
     metavar="ATTRIBUTE",
     nargs=1,
@@ -166,6 +185,7 @@ CLEAR_OPTION = click.option(
 )
 WIPE_OPTION = click.option(
     "--wipe",
+    "-X",
     help="Wipe all metadata attributes from FILE.",
     is_flag=True,
     default=False,
@@ -173,6 +193,7 @@ WIPE_OPTION = click.option(
 )
 APPEND_OPTION = click.option(
     "--append",
+    "-a",
     metavar="ATTRIBUTE VALUE",
     help="Append VALUE to ATTRIBUTE.",
     nargs=2,
@@ -181,6 +202,7 @@ APPEND_OPTION = click.option(
 )
 UPDATE_OPTION = click.option(
     "--update",
+    "-u",
     metavar="ATTRIBUTE VALUE",
     help="Update ATTRIBUTE with VALUE; for multi-valued attributes, "
     "this adds VALUE to the attribute if not already in the list.",
@@ -190,6 +212,7 @@ UPDATE_OPTION = click.option(
 )
 REMOVE_OPTION = click.option(
     "--remove",
+    "-r",
     metavar="ATTRIBUTE VALUE",
     help="Remove VALUE from ATTRIBUTE; only applies to multi-valued attributes.",
     nargs=2,
@@ -198,6 +221,7 @@ REMOVE_OPTION = click.option(
 )
 MIRROR_OPTION = click.option(
     "--mirror",
+    "-m",
     metavar="ATTRIBUTE1 ATTRIBUTE2",
     help="Mirror values between ATTRIBUTE1 and ATTRIBUTE2 so that ATTRIBUTE1 = ATTRIBUTE2; "
     "for multi-valued attributes, merges values; for string attributes, sets ATTRIBUTE1 = ATTRIBUTE2 "
@@ -209,14 +233,17 @@ MIRROR_OPTION = click.option(
 )
 BACKUP_OPTION = click.option(
     "--backup",
+    "-B",
     help="Backup FILE attributes.  "
-    "Backup file '.osxmetadata.json' will be created in same folder as FILE.",
+    "Backup file '.osxmetadata.json' will be created in same folder as FILE. "
+    "Only backs up attributes known to osxmetadata.",
     is_flag=True,
     required=False,
     default=False,
 )
 RESTORE_OPTION = click.option(
     "--restore",
+    "-R",
     help="Restore FILE attributes from backup file.  "
     "Restore will look for backup file '.osxmetadata.json' in same folder as FILE.",
     is_flag=True,
@@ -231,6 +258,35 @@ VERBOSE_OPTION = click.option(
     default=False,
     required=False,
 )
+COPY_FROM_OPTION = click.option(
+    "--copyfrom",
+    "-f",
+    metavar="SOURCE_FILE",
+    help="Copy attributes from file SOURCE_FILE.",
+    type=click.Path(exists=True),
+    nargs=1,
+    multiple=False,
+    required=False,
+)
+FILES_ONLY_OPTION = click.option(
+    "--files-only",
+    help="Do not apply metadata commands to directories themselves, only files in a directory.",
+    is_flag=True,
+    default=False,
+    required=False,
+)
+PATTERN_OPTION = click.option(
+    "--pattern",
+    "-p",
+    metavar="PATTERN",
+    help="Only process files matching PATTERN; only applies to --walk. "
+    "If specified, only files matching PATTERN will be processed as each directory is walked. "
+    "May be used for than once to specify multiple patterns. "
+    "For example, tag all *.pdf files in projectdir and subfolders with tag 'project': "
+    "osxmetadata --append tags 'project' --walk projectdir/ --pattern '*.pdf'",
+    multiple=True,
+    required=False,
+)
 
 
 @click.command(cls=MyClickCommand)
@@ -252,6 +308,9 @@ VERBOSE_OPTION = click.option(
 @BACKUP_OPTION
 @RESTORE_OPTION
 @VERBOSE_OPTION
+@COPY_FROM_OPTION
+@FILES_ONLY_OPTION
+@PATTERN_OPTION
 @click.pass_context
 def cli(
     ctx,
@@ -272,6 +331,9 @@ def cli(
     backup,
     restore,
     verbose,
+    copyfrom,
+    files_only,
+    pattern,
 ):
     """ Read/write metadata from file(s). """
 
@@ -331,6 +393,17 @@ def cli(
             attr1, attr2 = item
             attribute1 = ATTRIBUTES[attr1]
             attribute2 = ATTRIBUTES[attr2]
+
+            # avoid self mirroring
+            if attribute1 == attribute2:
+                click.echo(
+                    f"cannot mirror the same attribute: {attribute1.name} {attribute2.name}",
+                    err=True,
+                )
+                ctx.get_help()
+                ctx.exit(2)
+
+            # check type compatibility
             if (
                 attribute1.list != attribute2.list
                 or attribute1.type_ != attribute2.type_
@@ -345,98 +418,10 @@ def cli(
 
     # loop through each file, process it, then do backup or restore if needed
     for filename in files:
-        if walk and os.path.isdir(filename):
-            for root, _, filenames in os.walk(filename):
-                backup_file = pathlib.Path(root) / _BACKUP_FILENAME
-                if verbose:
-                    click.echo(f"Processing directory {root}")
-                if restore:
-                    try:
-                        backup_data = load_backup_file(backup_file)
-                    except FileNotFoundError:
-                        click.echo(
-                            f"Missing backup file {backup_file} for {root}, skipping restore",
-                            err=True,
-                        )
-                        backup_data = {}
-                else:
-                    backup_data = {}
-
-                for fname in filenames:
-                    fpath = pathlib.Path(f"{root}/{fname}").resolve()
-                    if restore and backup_data:
-                        try:
-                            attr_dict = backup_data[fname]
-                            if verbose:
-                                click.echo(f"    Restoring attribute data for {fpath}")
-                            md = osxmetadata.OSXMetaData(fpath)
-                            md._restore_attributes(attr_dict)
-                        except:
-                            if verbose:
-                                click.echo(
-                                    f"    Skipping restore for file {fpath}: not in backup file"
-                                )
-
-                    if verbose:
-                        click.echo(f"  Processing file: {fpath}")
-
-                    process_file(
-                        fpath,
-                        json_,
-                        set_,
-                        append,
-                        update,
-                        remove,
-                        clear,
-                        get,
-                        list_,
-                        mirror,
-                        wipe,
-                        verbose,
-                    )
-
-                    if backup:
-                        if verbose:
-                            click.echo(f"    Backing up attribute data for {fpath}")
-                        json_data = osxmetadata.OSXMetaData(fpath)._to_dict()
-                        backup_data[fpath.name] = json_data
-                if backup:
-                    # done walking through files in this folder, write the backup data
-                    write_backup_file(backup_file, backup_data)
-        elif os.path.isdir(filename):
-            # skip directory
-            if verbose:
-                click.echo(
-                    f"skipping directory: {filename}; use --walk to process directories"
-                )
-            continue
-        else:
-            fpath = pathlib.Path(filename).resolve()
-            backup_file = pathlib.Path(pathlib.Path(filename).parent) / _BACKUP_FILENAME
-            if restore:
-                try:
-                    backup_data = load_backup_file(backup_file)
-                    attr_dict = backup_data[pathlib.Path(fpath).name]
-                    if verbose:
-                        click.echo(f"  Restoring attribute data for {fpath}")
-                    md = osxmetadata.OSXMetaData(fpath)
-                    md._restore_attributes(attr_dict)
-                except FileNotFoundError:
-                    click.echo(
-                        f"Missing backup file {backup_file} for {fpath}, skipping restore",
-                        err=True,
-                    )
-                except KeyError:
-                    if verbose:
-                        click.echo(
-                            f"      Skipping restore for file {fpath}: not in backup file"
-                        )
-
-            if verbose:
-                click.echo(f"Processing file: {fpath}")
-
-            process_file(
-                fpath,
+        if not all([os.path.isdir(filename), walk, pattern]):
+            process_files(
+                ctx,
+                [filename],
                 json_,
                 set_,
                 append,
@@ -448,27 +433,154 @@ def cli(
                 mirror,
                 wipe,
                 verbose,
+                copyfrom,
+                backup,
+                restore,
+                walk,
+                files_only,
             )
 
-            if backup:
-                if verbose:
-                    click.echo(f"  Backing up attribute data for {fpath}")
-                # load the file if it exists, merge new data, then write out the file again
-                if backup_file.is_file():
-                    backup_data = load_backup_file(backup_file)
+        if walk and os.path.isdir(filename):
+            for root, dirnames, filenames in os.walk(filename):
+                if pattern:
+                    # only process files matching pattern
+                    filepaths = []
+                    for dirname in dirnames:
+                        for matches in [
+                            glob.glob(os.path.join(os.path.join(root, dirname), pat))
+                            for pat in pattern
+                        ]:
+                            filepaths.extend(matches)
                 else:
-                    backup_data = {}
-                json_dict = osxmetadata.OSXMetaData(fpath)._to_dict()
-                backup_data[pathlib.Path(fpath).name] = json_dict
-                write_backup_file(backup_file, backup_data)
+                    filepaths = [
+                        os.path.join(root, fname) for fname in dirnames + filenames
+                    ]
+                process_files(
+                    ctx,
+                    filepaths,
+                    json_,
+                    set_,
+                    append,
+                    update,
+                    remove,
+                    clear,
+                    get,
+                    list_,
+                    mirror,
+                    wipe,
+                    verbose,
+                    copyfrom,
+                    backup,
+                    restore,
+                    walk,
+                    files_only,
+                )
 
 
-def process_file(
-    fpath, json_, set_, append, update, remove, clear, get, list_, mirror, wipe, verbose
+def process_files(
+    ctx,
+    files,
+    json_,
+    set_,
+    append,
+    update,
+    remove,
+    clear,
+    get,
+    list_,
+    mirror,
+    wipe,
+    verbose,
+    copyfrom,
+    backup,
+    restore,
+    walk,
+    files_only,
+):
+    """ process list of files, calls process_single_file to process each file
+        options processed in this order: wipe, copyfrom, clear, set, append, remove, mirror, get, list
+        Note: expects all attributes passed in parameters to be validated as valid attributes 
+    """
+    for filename in files:
+        fpath = pathlib.Path(filename).resolve()
+        backup_file = pathlib.Path(pathlib.Path(filename).parent) / _BACKUP_FILENAME
+
+        if files_only and fpath.is_dir():
+            if verbose:
+                click.echo(f"Skipping directory: {fpath}")
+            continue
+
+        if verbose:
+            click.echo(f"Processing file: {fpath}")
+
+        if restore:
+            try:
+                backup_data = load_backup_file(backup_file)
+                attr_dict = backup_data[pathlib.Path(fpath).name]
+                if verbose:
+                    click.echo(f"  Restoring attribute data for {fpath}")
+                md = osxmetadata.OSXMetaData(fpath)
+                md._restore_attributes(attr_dict)
+            except FileNotFoundError:
+                click.echo(
+                    f"Missing backup file {backup_file} for {fpath}, skipping restore",
+                    err=True,
+                )
+            except KeyError:
+                if verbose:
+                    click.echo(
+                        f"  Skipping restore for file {fpath}: not in backup file"
+                    )
+
+        process_single_file(
+            ctx,
+            fpath,
+            json_,
+            set_,
+            append,
+            update,
+            remove,
+            clear,
+            get,
+            list_,
+            mirror,
+            wipe,
+            verbose,
+            copyfrom,
+        )
+
+        if backup:
+            if verbose:
+                click.echo(f"  Backing up attribute data for {fpath}")
+            # load the file if it exists, merge new data, then write out the file again
+            if backup_file.is_file():
+                backup_data = load_backup_file(backup_file)
+            else:
+                backup_data = {}
+            json_dict = osxmetadata.OSXMetaData(fpath).asdict()
+            backup_data[pathlib.Path(fpath).name] = json_dict
+            write_backup_file(backup_file, backup_data)
+
+
+def process_single_file(
+    ctx,
+    fpath,
+    json_,
+    set_,
+    append,
+    update,
+    remove,
+    clear,
+    get,
+    list_,
+    mirror,
+    wipe,
+    verbose,
+    copyfrom,
 ):
     """ process a single file to apply the options 
-        options processed in this order: wipe, clear, set, append, remove, mirror, get, list
-        Note: expects all attributes passed in parameters to be validated """
+        options processed in this order: wipe, copyfrom, clear, set, append, remove, mirror, get, list
+        Note: expects all attributes passed in parameters to be validated as valid attributes """
 
     logging.debug(f"process_file: {fpath}")
 
@@ -478,6 +590,8 @@ def process_file(
         attr_list = md.list_metadata()
         if verbose and attr_list:
             click.echo(f"Wiping metadata from {fpath}")
+        elif verbose:
+            click.echo(f"No metadata to wipe from {fpath}")
         for attr in attr_list:
             try:
                 attribute = ATTRIBUTES[attr]
@@ -490,11 +604,23 @@ def process_file(
                         f"  Unknown attribute {attr} on {fpath}, skipping", err=True
                     )
 
+    if copyfrom:
+        if verbose:
+            click.echo(f"Copying attributes from {copyfrom}")
+        src_md = osxmetadata.OSXMetaData(copyfrom)
+        for attr in src_md.list_metadata():
+            if verbose:
+                click.echo(f"  Copying {attr}")
+            md.set_attribute(attr, src_md.get_attribute(attr))
+
     if clear:
         for attr in clear:
             attribute = ATTRIBUTES[attr]
-            logging.debug(f"clearing {attr}")
+            if verbose:
+                click.echo(f"Clearing {attr}")
             md.clear_attribute(attribute.name)
+            if attr == "tags":
+                pass
 
     if set_:
         # set data
@@ -503,16 +629,31 @@ def process_file(
         for item in set_:
             attr, val = item
             attribute = ATTRIBUTES[attr]
-            logging.debug(f"setting {attr}={val}")
+
+            if attr in [*_TAGS_NAMES, *_FINDERINFO_NAMES]:
+                val = tag_factory(val)
+            if verbose:
+                click.echo(f"Setting {attr}={val}")
             try:
                 attr_dict[attribute].append(val)
             except KeyError:
                 attr_dict[attribute] = [val]
 
         for attribute, value in attr_dict.items():
-            logging.debug(f"value: {value}")
-            value = validate_attribute_value(attribute, value)
-            md.set_attribute(attribute.name, value)
+            if attribute.list:
+                # attribute expects a list so pass value (which is a list)
+                md.set_attribute(attribute.name, value)
+            else:
+                if len(value) == 1:
+                    # expected one and got one
+                    md.set_attribute(attribute.name, value[0])
+                else:
+                    click.echo(
+                        f"attribute {attribute.name} expects only a single value but {len(value)} provided",
+                        err=True,
+                    )
+                    ctx.get_help()
+                    ctx.exit(2)
 
     if append:
         # append data
@@ -521,14 +662,26 @@ def process_file(
         for item in append:
             attr, val = item
             attribute = ATTRIBUTES[attr]
-            logging.debug(f"appending {attr}={val}")
+
+            if not attribute.append:
+                click.echo(
+                    f"append is not a valid operation for attribute {attribute.name}",
+                    err=True,
+                )
+                ctx.get_help()
+                ctx.exit(2)
+
+            if attr in [*_TAGS_NAMES, *_FINDERINFO_NAMES]:
+                val = tag_factory(val)
+
+            if verbose:
+                click.echo(f"Appending {attr}={val}")
             try:
                 attr_dict[attribute].append(val)
             except KeyError:
                 attr_dict[attribute] = [val]
 
         for attribute, value in attr_dict.items():
-            value = validate_attribute_value(attribute, value)
             md.append_attribute(attribute.name, value)
 
     if update:
@@ -538,23 +691,46 @@ def process_file(
         for item in update:
             attr, val = item
             attribute = ATTRIBUTES[attr]
-            logging.debug(f"appending {attr}={val}")
+
+            if not attribute.update:
+                click.echo(
+                    f"update is not a valid operation for attribute {attribute.name}",
+                    err=True,
+                )
+                ctx.get_help()
+                ctx.exit(2)
+
+            if attr in [*_TAGS_NAMES, *_FINDERINFO_NAMES]:
+                val = tag_factory(val)
+
+            if verbose:
+                click.echo(f"Updating {attr}={val}")
             try:
                 attr_dict[attribute].append(val)
             except KeyError:
                 attr_dict[attribute] = [val]
 
         for attribute, value in attr_dict.items():
-            value = validate_attribute_value(attribute, value)
             md.update_attribute(attribute.name, value)
 
     if remove:
         # remove value from attribute
         # actually implemented with discard so no error raised if not present
-        # todo: catch errors and display help
         for attr, val in remove:
+            attribute = ATTRIBUTES[attr]
+            if not attribute.list:
+                click.echo(
+                    f"remove is not a valid operation for single-value attributes",
+                    err=True,
+                )
+                ctx.get_help()
+                ctx.exit(2)
+
+            if attr in [*_TAGS_NAMES, *_FINDERINFO_NAMES]:
+                val = tag_factory(val)
+            if verbose:
+                click.echo(f"Removing {attr}")
             try:
-                attribute = ATTRIBUTES[attr]
                 md.discard_attribute(attribute.name, val)
             except KeyError as e:
                 raise e
@@ -565,10 +741,24 @@ def process_file(
             # validation that attributes are compatible
             # will have occured prior to call to process_file
             attr1, attr2 = item
-            click.echo(f"Mirroring {attr1} {attr2}")
-            attribute1 = ATTRIBUTES[attr1]
+            if verbose:
+                click.echo(f"Mirroring {attr1} {attr2}")
 
-            if attribute1.list:
+            attribute1 = ATTRIBUTES[attr1]
+            attribute2 = ATTRIBUTES[attr2]
+
+            if attribute1.name == "tags":
+                tags = md.get_attribute(attr1)
+                attr1_values = [tag.name for tag in tags]
+                md.update_attribute(attr2, attr1_values)
+                md.set_attribute(attr1, [Tag(val) for val in md.get_attribute(attr2)])
+            elif attribute2.name == "tags":
+                attr1_values = md.get_attribute(attr1)
+                md.update_attribute(attr2, [Tag(val) for val in attr1_values])
+                tags = md.get_attribute(attr2)
+                attr2_values = [tag.name for tag in tags]
+                md.set_attribute(attr1, attr2_values)
+            elif attribute1.list:
                 # merge the two lists, assume attribute2 also a list due to
                 # previous validation
                 # update attr2 with any new values from attr1
@@ -588,19 +778,30 @@ def process_file(
             data["_filename"] = fpath.name
         for attr in get:
             attribute = ATTRIBUTES[attr]
-            logging.debug(f"getting {attr}")
             if json_:
-                if attribute.type_ == datetime.datetime:
-                    # need to convert datetime.datetime to string to serialize
-                    value = md.get_attribute(attribute.name)
-                    if type(value) == list:
-                        value = [v.isoformat() for v in value]
+                try:
+                    if attribute.name == "tags":
+                        tags = md.get_attribute(attribute.name)
+                        value = [[tag.name, tag.color] for tag in tags]
+                        data[attribute.constant] = value
+                    elif attribute.name == "finderinfo":
+                        finderinfo = md.get_attribute(attribute.name)
+                        value = [finderinfo.name, finderinfo.color]
+                        data[attribute.constant] = value
+                    elif attribute.type_ == datetime.datetime:
+                        # need to convert datetime.datetime to string to serialize
+                        value = md.get_attribute(attribute.name)
+                        if type(value) == list:
+                            value = [v.isoformat() for v in value]
+                        else:
+                            value = value.isoformat()
+                        data[attribute.constant] = value
                     else:
-                        value = value.isoformat()
-                    data[attribute.constant] = value
-                else:
-                    # get raw value
-                    data[attribute.constant] = md.get_attribute(attribute.name)
+                        # get raw value
+                        data[attribute.constant] = md.get_attribute(attribute.name)
+                except KeyError:
+                    # unknown attribute, ignore it
+                    pass
             else:
                 value = md.get_attribute_str(attribute.name)
                 click.echo(
@@ -615,6 +816,7 @@ def process_file(
             json_str = md.to_json()
             click.echo(json_str)
         else:
+            click.echo(f"{fpath}:")
             attribute_list = md.list_metadata()
             for attr in attribute_list:
                 try:
